@@ -22,11 +22,12 @@ Informations
 import numpy, math, scipy, pandas
 import numpy as np
 import pandas as pd
+from scipy.stats import zscore
 
 
 from sklearn.base import BaseEstimator, TransformerMixin
 
-from IPython.display import clear_output
+# from IPython.display import clear_output
 from sklearn import preprocessing
 from sklearn.preprocessing import (
     # MinMaxScaler,
@@ -37,6 +38,8 @@ from sklearn.preprocessing import (
 )
 from sklearn.pipeline import Pipeline
 
+from scipy import stats
+
 from .metrics import eval_information_value
 
 
@@ -297,7 +300,7 @@ class Encoder(BaseEstimator, TransformerMixin):
 
 
 class OneHotMissingEncoder(BaseEstimator, TransformerMixin):
-    """  """
+    """ """
 
     def __init__(self, columns, suffix="nan", sep="_", dummy_na=True, drop_last=False):
         """ """
@@ -313,7 +316,7 @@ class OneHotMissingEncoder(BaseEstimator, TransformerMixin):
         self.drop_last = drop_last
 
     def transform(self, X, **transform_params):
-        """    """
+        """ """
         X_copy = X.copy()
 
         final_columns = []
@@ -464,18 +467,37 @@ class ScalerDF(BaseEstimator, TransformerMixin):
         return X
 
 
+def _dataframe_transform(transformer, data):
+    if isinstance(data, (pd.DataFrame)):
+        return pd.DataFrame(
+            transformer.transform(data), columns=data.columns, index=data.index
+        )
+    else:
+        return transformer.transform(data)
+
+
 class MinMaxScaler(preprocessing.MinMaxScaler):
     def __init__(self, **kwargs):
-
         super().__init__(**kwargs)
 
-    def fit(self, X, y=None):
+    def transform(self, X):
+        return _dataframe_transform(super(), X)
 
-        super().fit(X)
+
+class StandardScaler(preprocessing.StandardScaler):
+    def __init__(self, **kwargs):
+        super().__init__(**kwargs)
 
     def transform(self, X):
+        return _dataframe_transform(super(), X)
 
-        return pd.DataFrame(super().transform(X), columns=X.columns, index=X.index)
+
+class RobustScaler(preprocessing.RobustScaler):
+    def __init__(self, **kwargs):
+        super().__init__(**kwargs)
+
+    def transform(self, X):
+        return _dataframe_transform(super(), X)
 
 
 class DataFrameImputer(TransformerMixin):
@@ -595,7 +617,7 @@ from sklearn.preprocessing import OrdinalEncoder
 
 
 class NumericBinner(BaseEstimator, TransformerMixin):
-    """  """
+    """ """
 
     def __init__(self, columns=None, min_bins=None, max_bins=None, n_bins=None):
 
@@ -617,8 +639,14 @@ class NumericBinner(BaseEstimator, TransformerMixin):
             self.min_bins = self.n_bins
             self.max_bins = self.n_bins
 
+        elif (self.min_bins is None) and (self.max_bins is not None):
+            self.min_bins = 2
+
+        elif (self.min_bins is not None) and (self.max_bins is None):
+            self.max_bins = 8
+
     def transform(self, X, **transform_params):
-        """    """
+        """ """
         X_copy = X.copy()
         for column_name in self.columns:
 
@@ -657,55 +685,84 @@ class NumericBinner(BaseEstimator, TransformerMixin):
                 X_local = X_copy.copy()
 
                 X_temp = X_local.loc[X_local[column_name].notna()][[column_name]].copy()
-
                 n = len(X_temp)
 
-                n_quant = 1000 if n >= 1000 else n
+                n_unique = len(X_temp[column_name].unique())
 
-                binner = Pipeline(
-                    steps=(
-                        ("scaler", RobustScaler()),
-                        (
-                            "quantile",
-                            QuantileTransformer(
-                                output_distribution="uniform", n_quantiles=n_quant
+                if n_unique >= i:
+
+                    n_quant = 1000 if n >= 1000 else n
+
+                    binner = Pipeline(
+                        steps=(
+                            ("scaler", RobustScaler()),
+                            (
+                                "quantile",
+                                QuantileTransformer(
+                                    output_distribution="uniform", n_quantiles=n_quant
+                                ),
                             ),
-                        ),
-                        (
-                            "binning",
-                            KBinsDiscretizer(
-                                n_bins=i, strategy="kmeans", encode="ordinal"
+                            (
+                                "binning",
+                                KBinsDiscretizer(
+                                    n_bins=i, strategy="kmeans", encode="ordinal"
+                                ),
                             ),
-                        ),
+                        )
                     )
-                )
 
-                X_temp = binner.fit_transform(X_temp)
+                    X_temp[column_name + "_transf"] = (
+                        binner.fit_transform(X_temp).flatten() + 1
+                    )
+                    X_temp[column_name + "_transf"] = X_temp[
+                        column_name + "_transf"
+                    ].astype(int, errors="ignore")
+                    X_temp.loc[X_copy[column_name].isna(), column_name + "_transf"] = 0
 
-                X_local.loc[X_local[column_name].notna(), column_name] = (
-                    X_temp.flatten() + 1
-                )
+                    summary = X_temp.groupby(column_name + "_transf").agg(
+                        min=(column_name, "min"),
+                        max=(column_name, "max"),
+                        mean=(column_name, "mean"),
+                        std=(column_name, "std"),
+                    )
 
-                X_local.loc[X_copy[column_name].isna(), column_name] = 0
+                    X_local.loc[X_local[column_name].notna(), column_name] = X_temp[
+                        column_name + "_transf"
+                    ]
 
-                X_local[column_name] = X_local[column_name].astype(int, errors="ignore")
+                    X_local.loc[X_copy[column_name].isna(), column_name] = 0
+                    X_local[column_name] = X_local[column_name].astype(
+                        int, errors="ignore"
+                    )
 
-                iv = eval_information_value(
-                    X_local[column_name], y, y_values=[0, 1], goods=0, treat_inf=True
-                )
+                    iv = eval_information_value(
+                        X_local[column_name],
+                        y,
+                        y_values=[0, 1],
+                        goods=0,
+                        treat_inf=True,
+                    )
 
-                X_local[column_name] = X_local[column_name].apply(
-                    lambda x: iv["woe"][x]
-                )
+                    iv = iv.merge(
+                        summary.reset_index(),
+                        how="left",
+                        right_on=column_name + "_transf",
+                        left_index=True,
+                    ).set_index(column_name + "_transf")
 
-                if (iv["iv"].sum() > iv_temp) & (
-                    np.isfinite(X_local[column_name].unique()).all()
-                ):
-                    iv_temp = iv["iv"].sum()
-                    bin_temp = iv
-                    binner_temp = binner
+                    X_local[column_name] = X_local[column_name].apply(
+                        lambda x: iv["woe"][x]
+                    )
+
+                    if (iv["iv"].sum() > iv_temp) & (
+                        np.isfinite(X_local[column_name].unique()).all()
+                    ):
+                        iv_temp = iv["iv"].sum()
+                        bin_temp = iv
+                        bin_temp.index.name = "bin"
+                        bin_temp = bin_temp.assign(total_iv=iv_temp)
+                        binner_temp = binner
 
-                clear_output()
             self.bins[column_name] = bin_temp
             self.binner[column_name] = binner_temp
 
@@ -714,3 +771,240 @@ class NumericBinner(BaseEstimator, TransformerMixin):
     def fit_transform(self, X, y):
         self.fit(X, y)
         return self.transform(X)
+
+
+class KS2Selector(TransformerMixin):
+    """ """
+
+    def __init__(self, alpha=0.05, invalid_label="OTHER", features=None, min_count=10):
+        self.invalid_label = invalid_label
+        self.alpha = alpha
+        self.features = features
+        self.relevant_classes = None
+        self.min_count = min_count
+
+    def fit(self, X, y):
+
+        classes_dict = {}
+        for col in self.features:
+
+            count_dict = X[col].value_counts().to_frame()
+
+            class_1st = count_dict[count_dict[col] >= self.min_count].index
+
+            list_ = []
+            for class_ in class_1st:
+                try:
+                    ks_res = stats.ks_2samp(y[X[col] == class_], y[X[col] != class_])
+                    list_.append((class_, *tuple(ks_res)))
+
+                except:
+                    pass
+
+            df_temp = pd.DataFrame(list_, columns=[col, "KS", "pval"])
+
+            relevant_classes = df_temp[df_temp["pval"] <= self.alpha][col]
+
+            classes_dict[col] = list(relevant_classes)
+
+        self.relevant_classes = classes_dict
+
+        return self
+
+    def transform(self, X, y=None):
+
+        for col in self.relevant_classes:
+            relevant_classes = self.relevant_classes[col]
+
+            X[col] = np.where(
+                np.isin(X[col], relevant_classes), X[col], self.invalid_label
+            )
+
+        return X
+
+    def fit_transform(self, X, y):
+        transformer = self.fit(X, y)
+        return self.transform(X)
+
+
+class Apply(BaseEstimator, TransformerMixin):
+    def __init__(self, function, features="all", axis=0, active=True):
+        self.features = features
+        self.function = function
+        self.axis = axis
+        self.active = active
+
+    def fit(self, X, y=None):
+        if type(self.features) == str:
+            if self.features == "all":
+                if self.axis == 0:
+                    self.features = list(X.columns)
+            else:
+                self.features = [self.features]
+
+        return self
+
+    def transform(self, X):
+        if not self.active:
+            return X
+        else:
+            return self.__transformation(X)
+
+    def fit_transform(self, X, y=None, **fit_params):
+        return super().fit_transform(X, y=y, **fit_params)
+
+    def __transformation(self, X_in):
+        X = X_in.copy()
+
+        if type(self.features) == str:
+            if self.features == "all":
+                if self.axis == 1:
+                    X = X.apply(self.function, axis=1)
+
+        if (type(self.features) == list) or (type(self.features) == tuple):
+            for col in self.features:
+                X[col] = X[col].apply(self.function)
+
+        return X
+
+
+class FillNaWithColumn(BaseEstimator, TransformerMixin):
+    def __init__(self, feature, column_source, active=True):
+        self.feature = feature
+        self.column_source = column_source
+        self.active = active
+
+    def fit(self, X, y=None):
+        return self
+
+    def transform(self, X):
+        X[self.feature] = X[self.feature].fillna(X[self.column_source])
+        return X
+
+    def fit_transform(self, X, y=None, **fit_params):
+        return super().fit_transform(X, y=y, **fit_params)
+
+
+class FillNaWithValue(BaseEstimator, TransformerMixin):
+    def __init__(self, feature, value, active=True):
+        self.feature = feature
+        self.value = value
+        self.active = active
+
+    def fit(self, X, y=None):
+        return self
+
+    def transform(self, X):
+        X[self.feature] = X[self.feature].fillna(self.value)
+        return X
+
+    def fit_transform(self, X, y=None, **fit_params):
+        return super().fit_transform(X, y=y, **fit_params)
+
+
+class StringReplace(BaseEstimator, TransformerMixin):
+    def __init__(self, features, pat, repl, n=-1, case=None, flags=0, regex=None):
+        self.features = features
+        self.pat = pat
+        self.repl = repl
+        self.n = n
+        self.case = case
+        self.flags = flags
+        self.regex = regex
+
+    def fit(self, X, y=None):
+        if type(self.features) == str:
+            self.features = [self.features]
+
+        return self
+
+    def transform(self, X):
+        for col in self.features:
+            X[col] = X[col].str.replace(
+                pat=self.pat,
+                repl=self.repl,
+                n=self.n,
+                case=self.case,
+                flags=self.flags,
+                regex=self.regex,
+            )
+        return X
+
+    def fit_transform(self, X, y=None, **fit_params):
+        return super().fit_transform(X, y=y, **fit_params)
+
+
+class Converter(BaseEstimator, TransformerMixin):
+    def __init__(self, features, type):
+        self.features = features
+        self.type = type
+
+    def fit(self, X, y=None):
+        if type(self.features) == str:
+            self.features = [self.features]
+        return self
+
+    def transform(self, X):
+        for col in self.features:
+            X[col] = X[col].astype(self.type, errors="ignore")
+        return X
+
+    def fit_transform(self, X, y=None, **fit_params):
+        return super().fit_transform(X, y=y, **fit_params)
+
+
+class OutliersRemover(BaseEstimator, TransformerMixin):
+    def __init__(self, features, z=3):
+        self.features = features
+        self.z = z
+
+    def fit(self, X, y=None):
+        if type(self.features) == str:
+            self.features = [self.features]
+        return self
+
+    def transform(self, X):
+        for col in self.features:
+            X[col] = self.removing_outliers(X, col, z=self.z)
+        return X
+
+    def fit_transform(self, X, y=None, **fit_params):
+        return super().fit_transform(X, y=y, **fit_params)
+
+    def removing_outliers(self, data, column, z=3):
+        X = data.copy()
+
+        X["z_score"] = zscore(X[column], nan_policy="omit")
+
+        list_ref = X[(X["z_score"] <= z) & (X["z_score"] >= -z)][column]
+
+        X.loc[X["z_score"] > z, column] = list_ref.max()
+        X.loc[X["z_score"] < -z, column] = list_ref.min()
+
+        return X[column]
+
+
+class CenteredMeanScaler(preprocessing.MinMaxScaler):
+    def __init__(self):
+        super().__init__()
+        self.mean = None
+
+    def fit(self, X, y=None):
+
+        X = X.copy()
+
+        super().fit(X, y)
+
+        self.mean = np.nanmean(super().transform(X))
+
+        return self
+
+    def transform(self, X, y=None):
+
+        X = X.copy()
+
+        X = super().transform(X)
+
+        S = (X - X * self.mean) / (X + self.mean * (1 - 2 * X))
+
+        return S
